﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using System.Xml.Serialization;

namespace rcon
{
    public partial class Client : Form
    {
        public Client(string _host, string _port, string _password)
        {
            InitializeComponent();

            host = _host;
            port = int.Parse(_port);
            password = _password;
        }

        int port;
        string host, password;
        SourceRcon.SourceRcon r;
        List<CVAR> cvars, subset;
        string hilighted;
        
        List<string> logmessages;
        bool printlog;
        bool dirty;

        private void Client_Load(object sender, EventArgs e)
        {
            dirty = false;
            printlog = true;
            logmessages = new List<string>();

            hilighted = "";
            cvars = new List<CVAR>();
            subset = new List<CVAR>();
            r = new SourceRcon.SourceRcon();
            r.Errors += new SourceRcon.StringOutput(err);
            r.ServerOutput += new SourceRcon.StringOutput(log);
            r.Connect(host, port, password);

            Timer t = new Timer();
            t.Tick += new EventHandler(t_Tick);
            t.Interval = 10;
            t.Start();

            string[] fonts = {
                "Consolas",
                "Lucida Console",
                "Liberation Mono"
            };
            string font = "";
            FontStyle fs = FontStyle.Regular;
            foreach (string f in fonts)
            {
                fs = hasFont(f);
                if (fs != FontStyle.Strikeout)
                {
                    font = f;
                    break;
                }
            }
            textBox2.Font = !string.IsNullOrEmpty(font) ?
                new System.Drawing.Font(font, textBox2.Font.SizeInPoints * 1.2f, fs) :
                new System.Drawing.Font(FontFamily.GenericMonospace, textBox2.Font.SizeInPoints);

            try
            {
                XmlSerializer x = new XmlSerializer(typeof(List<CVAR>));
                using (System.IO.FileStream s =
                    new System.IO.FileStream("rcon.cvars",
                        System.IO.FileMode.Open,
                        System.IO.FileAccess.Read))
                {
                    cvars = (List<CVAR>)x.Deserialize(s);
                    cvar2html();
                }
            }
            catch { cvars = new List<CVAR>(); }
        }

        FontStyle hasFont(string fontName)
        {
            if (hasFont(fontName, FontStyle.Regular)) return FontStyle.Regular;
            if (hasFont(fontName, FontStyle.Bold)) return FontStyle.Bold;
            if (hasFont(fontName, FontStyle.Italic)) return FontStyle.Italic;
            return FontStyle.Strikeout;
        }
        bool hasFont(string fontName, FontStyle style)
        {
            using (var test = new Font(fontName, 8, style))
            {
                return 0 == string.Compare(fontName, test.Name,
                    StringComparison.InvariantCultureIgnoreCase);
            }
        }

        void cvar2html()
        {
            try
            {
                StringBuilder sb = new StringBuilder();
                sb.AppendLine(@"<!DOCTYPE html><html lang=""en""><head>
	<meta charset=""utf-8"" />
	<title>cvar list</title>
	<meta http-equiv=""X-UA-Compatible"" content=""IE=edge"">
	<meta name=""viewport"" content=""width=device-width, initial-scale=0.5"">
	<style>
html, td {
	color: #fff;
	background: #2c2c2c;
	font-family: monospace, monospace;
}
table {
	border-collapse: collapse;
}
tr td {
	background: #303030;
	padding: .2em;
}
tr:nth-child(2n) td
{
	background: #282828;
}
tr:hover td {
	background: #222;
}
.def {
	color: #555;
}
</style></head><body><table>");
                foreach (CVAR cvar in cvars)
                    sb.AppendFormat("<tr><td>{0}</td><td{3}>{1}</td><td>{2}</td></tr>\r\n",
                        html(cvar.key), html(cvar.value), html(cvar.desc),
                        cvar.value == "cmd" ? " class=\"def\"" : "");

                sb.AppendLine("</table></body></html>");
                System.IO.File.WriteAllText("rcon-cvar-list.html", sb.ToString(), Encoding.UTF8);
            }
            catch { }
        }

        string html(string v)
        {
            return v.Replace("&", "&nbsp;").Replace("<", "&lt;").Replace(">", "&gt;");
        }

        void err(string msg)
        {
            log(msg);
        }

        void log(string msg)
        {
            lock (logmessages)
            {
                string[] lines = msg.Replace("\r", "").Split('\n');
                logmessages.AddRange(lines);
                while (logmessages.Count > 5000)
                {
                    logmessages.RemoveRange(0, 100);
                }
                dirty = true;
            }
        }

        void t_Tick(object sender, EventArgs e)
        {
            if (!dirty) return;
            if (!printlog) return;
            StringBuilder sb = new StringBuilder();
            lock (logmessages)
            {
                dirty = false;
                foreach (string str in logmessages)
                {
                    sb.AppendLine(str);
                }
            }
            textBox2.Text = sb.ToString();
            if (textBox2.Text.Length > 3)
            {
                textBox2.SelectionStart = textBox2.Text.Length;
                textBox2.ScrollToCaret();
            }
        }

        private void textBox1_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                e.Handled = true;
                r.ServerCommand(textBox1.Text);
                textBox1.SelectAll();
            }
            else if (e.KeyCode == Keys.Up)
            {
                e.Handled = true;
                shiftHilight(-1);
            }
            else if (e.KeyCode == Keys.Down)
            {
                e.Handled = true;
                shiftHilight(+1);
            }
            else if (e.KeyCode == Keys.Back)
            {
                e.Handled = true;
                int texlen = textBox1.Text.Length;
                int selofs = textBox1.SelectionStart;
                int sellen = textBox1.SelectionLength;
                if (sellen > 0)
                {
                    if (selofs > 0)
                        selofs--;
                    if (sellen < texlen)
                        sellen++;
                    textBox1.SelectionStart = selofs;
                    textBox1.SelectionLength = sellen;
                }
            }
        }

        private void button1_Click(object sender, EventArgs e)
        {
            lock (logmessages)
            {
                logmessages.Clear();
                dirty = true;
            }
        }

        private void button2_Click(object sender, EventArgs e)
        {
            button1_Click(sender, e);
            r.ServerCommand("cvarlist");
            bool gotit = false;
            printlog = false;
            button1.Enabled = false;
            button2.Enabled = false;
            textBox1.Enabled = false;
            Application.DoEvents();
            while (!gotit)
            {
                lock (logmessages)
                {
                    foreach (string str in logmessages)
                    {
                        if (str.Contains("total convars/concommands"))
                        {
                            gotit = true;
                            break;
                        }
                    }
                    if (gotit)
                    {
                        cvars.Clear();
                        foreach (string str in logmessages)
                        {
                            var re = Regex.Match(str, "^([^ ][^ ]*)  *: ([^: ]*)  *: [^:]* *(:.*)", RegexOptions.Singleline).Groups;
                            if (re.Count == 4)
                            {
                                cvars.Add(new CVAR(re[1].Value, re[2].Value, re[3].Value.Substring(1).Trim()));
                                //MessageBox.Show(cvars[cvars.Count - 1].ToString());
                            }
                        }
                        XmlSerializer x = new XmlSerializer(typeof(List<CVAR>));
                        using (var s = new System.IO.FileStream("rcon.cvars", System.IO.FileMode.Create))
                        {
                            x.Serialize(s, cvars);
                        }
                        cvar2html();
                    }
                }
            }
            button1_Click(sender, e);
            button1.Enabled = true;
            button2.Enabled = true;
            textBox1.Enabled = true;
            printlog = true;
        }

        private void textBox1_TextChanged(object sender, EventArgs e)
        {
            if (disregard) return;
            subset.Clear();
            hilighted = "";
            string t = textBox1.Text.ToLower(); //.Split(' ')[0];
            foreach (CVAR c in cvars)
            {
                if (c.key.ToLower().StartsWith(t))
                {
                    subset.Add(c);
                    if (subset.Count > 32)
                    {
                        break;
                    }
                    else if (subset.Count == 1)
                    {
                        hilighted = c.key;
                    }
                }
            }
            drawHints();
        }

        void shiftHilight(int direction)
        {
            int i = subset.FindIndex(w => w.key == hilighted);
            i += direction;
            if (i < 0)
            {
                i = subset.Count - 1;
            }
            if (i >= subset.Count)
            {
                i = 0;
            }
            hilighted = subset[i].key;
            drawHints();
        }

        bool disregard = false;
        void drawHints()
        {
            if (disregard) return;
            disregard = true;

            StringBuilder s = new StringBuilder();
            string userinput = textBox1.Text;
            if (textBox1.SelectionLength > 0)
            {
                userinput = userinput.Substring(0, textBox1.SelectionStart);
            }
            foreach (CVAR c in subset)
            {
                s.AppendLine(c.key + "   (" + c.value + ")");
                if (c.key == hilighted)
                {
                    //s.AppendLine();
                    s.AppendLine("    " + c.desc);
                    s.AppendLine();
                }
            }
            textBox2.Text = s.ToString();
            if (!string.IsNullOrEmpty(hilighted))
            {
                textBox1.Text = hilighted;
                textBox1.Select(userinput.Length, hilighted.Length);
            }
            disregard = false;
        }
    }
}
